import java.io.*;
import java.awt.*;
import VM.*;

public class Client4004 extends ClientBase implements GUI.Paintable
{
	// start or load --------------------------------------------------------------------

	public static void main(String[] arg) throws Exception 
//	  { (new File("vm4004.dump").exists() ? (Client4004)load(4004) : new Client4004()).run(); }
	  { new Client4004().run(); }
	public Client4004() throws Exception { super(4004); }

	// strategies -----------------------------------------------------------------------

	interface Strategy
	{
		public boolean doit() throws IOException;
	}

	class DoNothing implements Strategy, Serializable
	{
		public boolean doit() throws IOException { step(0, 0); return false; }
	}

	class TwoSteps implements Strategy, Serializable
	{
		public boolean doit() throws IOException { step(0, 0); step(0, 0); return true; }
	}

	class Manual implements Strategy, Serializable
	{
		double dvx, dvy; int T;
		Manual(double dvx, double dvy, int T) { this.dvx=dvx; this.dvy=dvy; this.T=T; }
		public boolean doit() throws IOException {
			if(vm.getTimeStamp()<T){step(0,0);return false;}
			else{ step(dvx, dvy); return true;}
		}
	}

	class SameSpeed implements Strategy, Serializable
	{
		int k, T;
		SameSpeed(int k, int T) { this.k=k; this.T=T; }
		public boolean doit() throws IOException
		{ if(vm.getTimeStamp()<T){step(0,0);return false;}
          else {step(VTX4[k]-VX, VTY4[k]-VY); return true;} }
	}

	class SpeedChange implements Strategy, Serializable
	{
		double p1; int T;
		SpeedChange(double p1, int T) { this.p1=p1; this.T=T; }
		public boolean doit() throws IOException
		{ if(vm.getTimeStamp()<T){step(0,0);return false;}
          else {step(VX*(p1-1), VY*(p1-1)); return true;} }
	}

	class GoToCircle implements Strategy, Serializable
	{
		int k; int T; double p1; boolean inverse;

		GoToCircle(int k, int T, double p1)
		  { this.k=k; this.T=T; this.p1=p1; this.inverse=false; }
		GoToCircle(int k, int T, double p1, boolean inverse)
		  { this.k=k; this.T=T; this.p1=p1; this.inverse=inverse; }

		double start_r=0, end_r=0;
		public boolean doit() throws IOException
		{
			if( vm.getTimeStamp() < T )
			{
				step(0, 0);
				return false;
			}
			else if( vm.getTimeStamp() == T )
			{
				start_r = SR;
				end_r   = TR4[k] + p1;
				double f = (Math.sqrt(end_r*2 / (end_r+SR)) - 1) * V;
				System.out.println("GoToCircle-"+k+": "+vm.getTimeStamp()+"  ("+N_VX*f+", "+N_VY*f+")");
				step(N_VX*f, N_VY*f);
				return false;
			}
			else
			{
				if( Math.abs(SR-start_r)>Math.abs(SR-end_r) && ((!inverse) == (PR>SR)) )
				{
					double v = Math.sqrt(G*ME / end_r);
					double f = (Math.sqrt(start_r*2 / (start_r+SR)) - 1) * -v;
					System.out.println("GoToCircle-"+k+": "+vm.getTimeStamp()+"  ("+N_VX*f+", "+N_VY*f+")");
					step(N_VX*f, N_VY*f);
					return true;
				}
				else
				{
					step(0, 0);
					return false;
				}
			}
		}
	}

	class StrikeOut implements Strategy, Serializable
	{
		int k; int T; double p1;

		StrikeOut(int k, int T, double p1)
		  { this.k=k; this.T=T; this.p1=p1; }

		double start_r=0, end_r=0;
		public boolean doit() throws IOException
		{
			if( vm.getTimeStamp() < T )
			{
				step(0, 0);
				return false;
			}
			else if( vm.getTimeStamp() == T )
			{
				System.out.println("StrikeOut-"+k+": "+vm.getTimeStamp());

				start_r = SR;
				end_r   = TR4[k] + p1;
				double f = (Math.sqrt(end_r*2 / (end_r+SR)) - 1) * V;
				step(N_VX*f, N_VY*f);
				return true;
			}
			else
			{
				step(0, 0);
				return false;
			}
		}
	}

	// main loop -------------------------------------------------------------------------

	transient Strategy[] Q;
	int Qi     = 0;
	int Target = 0;

	void run() throws IOException
	{
		if( Q == null )
			Q = new Strategy[]{
				new TwoSteps(),
				new Manual(-70.14521005073367, -2.9532728964829, 35),
				new Manual(2577.3, 0.2, 1992),
				new Manual(524.3, 0.1, 5635),
				new Manual(-20.2, 1500.39, 10696),
				new Manual(-980.1, 526.95, 17284),
				new Manual(-154.7, 0, 24940),
				new Manual(-248.83, 0, 34852),
				new Manual(-279.4, 0, 47229),
				new Manual(-175.5,-400.1,62557),
				new Manual(0.04,-194.86,79695),
				new Manual(0.1, -120.71, 99373),
				new Manual(872.4, 0, 122157),
				new Manual(-2.271, 0.0001, 300001),
				new SpeedChange(1.02, 421771),
				new Manual(0, 0, 421772),
			};

		double totalScore = 0;

		int FINAL_TARGET = 11;
		while( Qi < Q.length ) // Target <= FINAL_TARGET )
		{
			// State Machine...
			if( Qi >= Q.length )
				step(0, 0);
			else
				if( Q[Qi].doit() )
					if( ++Qi < Q.length )
						save();

			// Target Management
			if( SCORE < 0 )
			{
				System.out.println("CRASHED!");
				System.exit(0);
			}
			if( Target<=11 && SCORE4[Target] > 0 )
			{
				double sThis = 8*(75.0/12*(2000000-vm.getTimeStamp())/2000000);
				totalScore += sThis;
				System.out.println("!! CAPTURED  "+Target+" !! "+vm.getTimeStamp()+"s"+" (score "+sThis+") (fuel: "+FUEL/FUEL0+")");
				++Target;
			}

			// Zoom
			boolean ZOOM =
				Target <= 11
				&& Math.sqrt((TX4[Target]-SX)*(TX4[Target]-SX) + (TY4[Target]-SY)*(TY4[Target]-SY))<FROMME_RANGE
				&& Target>=0;//FINAL_TARGET;
			FROMME = ZOOM;

			FROMME_RANGE = 2000000;
			if(ZOOM) wait(2900000);
//			FROMME_RANGE = 200000;
//			if(ZOOM) wait(90000000);
//			FROMME_RANGE = 50000;
//			if(ZOOM) wait(900000000);


//			else {FROMME=true; FROMME_RANGE=500000000.0;}
		}

		System.out.println( "score: " + (totalScore + 200*(FUEL+FUELTANK)/(FUEL0+FUELTANK0)) );
		System.exit(0);
	}
}
