import java.io.*;
import java.awt.*;
import java.util.*;
import VM.*;

public class ClientBase implements GUI.Paintable, Serializable
{
	// API -----------------------------------------------------------------

	public VM vm;
		// Main Virtual Machine:
		//
		//   vm.step( deltaVx, deltaVy );
		//      DONT CALL THIS METHOD DIRECTLY. use ClinentBase#step
		//
		//   double v = vm.readPort(n);
		//      read output port n
		//
		//   VM vm2 = vm.snapshot();
		//      create current snapshot

	public int MODE;

	public static final double G  = 6.67428e-11;
	public static final double ME = 6e+24;
	public static final double ER = 6.357e+6;
	public static final double MMOON = 7.347e+22;
	public static final double RMOON = 1.738e+6;


	public double SCORE;      // 1-
	public double  FUEL;      // 1-
	public double FUEL0 = -1; // 1-: initial fuel
	public double    SX;      // 1-: absolute coordinate of MySelf:  earth=(0,0)
	public double    SY;      // 1-: absolute coordinate of MySelf:  earth=(0,0)
	public double    SR;      // 1-: sqrt(TX^2 + TY^2);
	public double    SG;      // 1-: currnet gravity accel
	public double    GX;      // 1-: currnet gravity accel
	public double    GY;      // 1-: currnet gravity accel
	public double  N_GX;      // 1-: currnet gravity accel
	public double  N_GY;      // 1-: currnet gravity accel
	public double  N_SX; // 1-: normalized
	public double  N_SY; // 1-: normalized
	public double    TX;      // 1-: absolute coorinate of Target:  earth=(0,0)
	public double    TY;      // 1-: absolute coorinate of Target:  earth=(0,0)
	public double    TR;      // 1-: sqrt(TX^2 + TY^2)
	public double    PX; // 2-: prev_sx
	public double    PY; // 2-: prev_sy
	public double    PR; // 2-: prev_sr
	public double    VX; // 2-: velocity
	public double    VY; // 2-: velocity
	public double     V; // 2-: |velocity|
	public double  N_VX; // 2-: normalized
	public double  N_VY; // 2-: normalized
	public double   PTX; // 2-: prev_tx 
	public double   PTY; // 2-: prev_ty
	public double   PTR; // 2-: prev_ty
	public double   VTX; // 2-: velocity of T
	public double   VTY; // 2-: velocity of T
	public double    VT; // 2-: |velocity| of T
	public double  N_VTX; // 2-: normalized of T
	public double  N_VTY; // 2-: normalized of T

	public double FUELTANK;
	public double FUELTANK0 = -1;
	public double SCORE4[] = new double[12]; // 0 is not used
	public double TX4[] = new double[12]; // 0 for fuel station, 1-12: target
	public double TY4[] = new double[12];
	public double TR4[] = new double[12];
	public double PTX4[] = new double[12];
	public double PTY4[] = new double[12];
	public double PTR4[] = new double[12];
	public double VTX4[] = new double[12];
	public double VTY4[] = new double[12];
	public double VT4[] = new double[12];
	public double N_VTX4[] = new double[12];
	public double N_VTY4[] = new double[12];
	public double MOON_X;
	public double MOON_Y;
	public double MOON_R;

	public synchronized void save() throws IOException
	{
		ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("vm"+vm.getConfigID()+".dump"));
		oos.writeObject(this);
	}
	static public ClientBase load(int config) throws IOException, ClassNotFoundException
	{
		ObjectInputStream ois = new ObjectInputStream(new FileInputStream("vm"+config+".dump"));
		ClientBase cb = (ClientBase)ois.readObject();
		cb.startGUI();

		File fp = new File(config+".osf");
		if( fp.exists() ) {
			Runtime.getRuntime().exec("cp "+config+".osf "+config+".bak.osf" );
			fp.delete();
		}
		return cb;
	}

	public synchronized void step( double dvx, double dvy ) throws IOException
	{
		PX = SX;
		PY = SY;
		PR = SR;
		PTX = TX;
		PTY = TY;
		PTR = TR;
		if(MODE==4) {
			for(int k=0; k<12; ++k) {
				PTX4[k] = TX4[k];
				PTY4[k] = TY4[k];
				PTR4[k] = TR4[k];
			}
		}

		vm.step(dvx, dvy);

		SCORE = vm.readPort(0);
		FUEL  = vm.readPort(1);
		if(FUEL0<0) FUEL0 = FUEL;
		SX    = -vm.readPort(2);
		SY    = -vm.readPort(3);
		SR    = Math.sqrt(SX*SX+SY*SY);
		N_SX  = SX/SR;
		N_SY  = SY/SR;
		SG    = G * ME / (SR*SR);
		N_GX  = -N_SX;
		N_GY  = -N_SY;
		GX    = N_GX * SG;
		GY    = N_GY * SG;
		VX    = SX - PX;
		VY    = SY - PY;
		V     = Math.sqrt(VX*VX+VY*VY);
		N_VX  = V==0 ? 0 : VX/V;
		N_VY  = V==0 ? 0 : VY/V;

		if(MODE==1)
			TR = vm.readPort(4);
		if(MODE==2 || MODE==3)
		{
			TX = SX+vm.readPort(4);
			TY = SY+vm.readPort(5);
			TR = Math.sqrt(TX*TX+TY*TY);
			VTX    = TX - PTX;
			VTY    = TY - PTY;
			VT     = Math.sqrt(VTX*VTX+VTY*VTY);
			N_VTX  = VT==0 ? 0 : VTX/VT;
			N_VTY  = VT==0 ? 0 : VTY/VT;
		}
		if(MODE==4)
		{
			FUELTANK = vm.readPort(6);
			if(FUELTANK0<0) FUELTANK0=FUELTANK;
			MOON_X = SX+vm.readPort(100);
			MOON_Y = SY+vm.readPort(101);
			MOON_R = Math.sqrt(MOON_X*MOON_X + MOON_Y*MOON_Y);
			for(int k=0; k<12; ++k) {
				TX4[k] = SX + vm.readPort(3*(k-1)+7);
				TY4[k] = SY + vm.readPort(3*(k-1)+8);
				if(k>0) SCORE4[k] = vm.readPort(3*(k-1)+9);
				else {
					if(SCORE4[k]==0 && Math.sqrt(
						(SX-TX4[k])*(SX-TX4[k])+(SY-TY4[k])*(SY-TY4[k]))<1000 )
						SCORE4[k] = 1.0;
				}
				TR4[k] = Math.sqrt(TX4[k]*TX4[k]+TY4[k]*TY4[k]);
				VTX4[k] = TX4[k]-PTX4[k];
				VTY4[k] = TY4[k]-PTY4[k];
				VT4[k] = Math.sqrt(VTX4[k]*VTX4[k]+VTY4[k]*VTY4[k]);
				N_VTX4[k] = VTX4[k]/VT4[k];
				N_VTY4[k] = VTY4[k]/VT4[k];
			}
		}
	}


	public void wait(int n)
	{
		for(int zz=0; zz<n; ++zz) ;
	}

	private int score(int n)
	{
		int ret = 0;
		while( n > 0 )
		{
			n /= 2;
			++ret;
		}
		return ret;
	}

	public void endgame_check(int t)
	{
		if( SCORE != 0 )
		{
			System.out.println("time : " + t + " (" + (30 - score(t / 1000)) + ")");
			System.out.println("fuel : " + FUEL + "/" + FUEL0 + " (" + (45 * (FUEL0 - FUEL) / FUEL0) + ")");
			System.out.println("myscore : " + (25 + 45*(FUEL0-FUEL)/FUEL0 + 30-score(t / 1000)));
			System.out.println(SCORE<0 ? "GAME OVER" : "CONGRATULATIONS!");
			System.out.println("  score : " + SCORE);
			System.exit(0);
		}
	}

	// implmentation -------------------------------------------------------

	private Font font;

	private void startGUI() {
		font = new Font("MS Gothic", Font.PLAIN, 24);
		new GUI(this);
	}

	public ClientBase( int config ) throws Exception
	{
		this.MODE = config/1000;

		VM raw_vm =
			MODE==1 ? new VM1(config) :
			MODE==2 ? new VM2(config) :
			MODE==3 ? new VM3(config) :
			MODE==4 ? new VM4(config) : null;
		this.vm = new LoggingVM(raw_vm, config+".osf");
		startGUI();
	}

	public ClientBase( int config, boolean use_log ) throws Exception
	{
		this.MODE = config/1000;

		VM raw_vm =
			MODE==1 ? new VM1(config) :
			MODE==2 ? new VM2(config) :
			MODE==3 ? new VM3(config) :
			MODE==4 ? new VM4(config) : null;
		if( use_log )
			this.vm = new LoggingVM(raw_vm, config+".osf");
		else
			this.vm = raw_vm;
		startGUI();
	}

	private java.util.List<Double> H_TX = new LinkedList<Double>();
	private java.util.List<Double> H_TY = new LinkedList<Double>();
	public boolean FROMME = false;
	public double FROMME_RANGE = 1000000;
	public double MAGNIFY = 10;

	// zahyo- henkan
	public double tr=1, R=250;
	public void drawOval(Graphics g, double x, double y, double w, double h)
	  { if(FROMME){x-=SX; y-=SY;}
	    double ra = R / (FROMME ? FROMME_RANGE : tr);
	    g.drawOval( (int)(300+x*ra),(int)(300-(y+h)*ra), (int)(w*ra), (int)(h*ra) ); }
	public void fillOval(Graphics g, double x, double y, double w, double h)
	  { if(FROMME){x-=SX; y-=SY; }
	    double ra = R / (FROMME ? FROMME_RANGE : tr);
	    g.fillOval( (int)(300+x*ra),(int)(300-(y+h)*ra), (int)(w*ra), (int)(h*ra) ); }
	public void drawDot(Graphics g, double x, double y, int sz)
	  { if(FROMME){x-=SX; y-=SY; }
	    double ra = R / (FROMME ? FROMME_RANGE : tr);
	    g.fillOval( (int)(300+x*ra)-sz/2,(int)(300-y*ra)-sz/2, sz, sz ); }
	public void drawString(Graphics g, String s, double x, double y, int sz)
	  { if(FROMME){x-=SX; y-=SY; }
	    double ra = R / (FROMME ? FROMME_RANGE : tr);
	    g.drawString( s, (int)(300+x*ra)-sz/2,(int)(300-y*ra)+sz/2 ); }

	public synchronized void paint( Graphics g )
	{
		g.setFont(font);
		Color ColorEarth  = new Color(200,255,255);
		Color ColorMoon  = new Color(200,200,0);
		Color ColorTarget = Color.orange;
		Color ColorFar  = Color.green;
		Color ColorNear = Color.blue;
		Color ColorGood = Color.red;
		Color ColorGone = Color.black;

		tr = Math.max(tr, SR);
		if( MODE<=3 )
			tr = Math.max(tr, TR);
		else if( MODE==4 )
		{
			for(int k=0; k<12; ++k)
				tr = Math.max(tr, TR4[k]/MAGNIFY);
		}
		double ra = R / (FROMME ? FROMME_RANGE : tr);

		// Time & Fuel
		g.setColor(Color.gray);
		g.drawString(vm.getTimeStamp()+" sec", 460, 550);
		if( MODE<=3 )
			g.drawString("F:"+(int)(FUEL/FUEL0*100)+" % (of "+(int)FUEL0+")", 0, 550);
		else
			g.drawString("F:"+(int)(FUEL/FUEL0*100)+" % (of "+(int)FUEL0+") + "+(vm.readPort(6)), 0, 550);

		// Earth
		g.setColor(ColorEarth);
		fillOval(g, -ER, -ER, 2*ER, 2*ER);
		if(MODE==4) {
			g.setColor(ColorMoon);
			fillOval(g, MOON_X-RMOON, MOON_Y-RMOON, 2*RMOON, 2*RMOON);
		}

		// Target
		g.setColor(ColorTarget);
		if(MODE<=2)
			drawOval(g, -TR, -TR, 2*TR, 2*TR );
		if(MODE==2 || MODE==3)
			drawDot(g, TX, TY, 10);
		if(MODE==3) {
			for(int i=0; i<H_TX.size(); ++i)
				drawDot(g, H_TX.get(i), H_TY.get(i), 4);
			H_TX.add(TX);
			H_TY.add(TY);
			if(200 < H_TX.size())
			{
				int i = (int)(Math.random()*H_TX.size());
				H_TX.remove(i);
				H_TY.remove(i);
			}
		}
		if(MODE==4) {
			for(int i=0; i<H_TX.size(); ++i)
				drawDot(g, H_TX.get(i), H_TY.get(i), 4);
			for(int k=0; k<12; ++k) {
				if( k>0 ) {
					g.setColor(ColorTarget);
					if( k>0 && SCORE4[k] == 0 )
					{
						//drawDot(g, TX4[k], TY4[k], 10);
						drawString(g, ""+k, TX4[k], TY4[k], 12);

						H_TX.add(TX4[k]);
						H_TY.add(TY4[k]);
						if(400 < H_TX.size())
						{
							int i = (int)(Math.random()*H_TX.size());
							H_TX.remove(i);
							H_TY.remove(i);
						}
					}
				} else {
					g.setColor(new Color(255,0,255));
					drawDot(g, TX4[k], TY4[k], 10);
				}
			}
		}

		// MySelf
		double NTR = 0, NTX=TX, NTY=TY;
		if( MODE<=3 ) NTR = TR;
		else {
			for(int k=0; k<12; ++k) {
				if( Math.abs(TR4[k]-SR) < Math.abs(NTR-SR) ) {
					NTR = TR4[k];
					NTX = TX4[k];
					NTY = TY4[k];
				}
			}
		}

		if( MODE<=2 ) {
			if( NTR-SR<-1000 )         g.setColor(ColorGone);
			else if( NTR-SR<+1000 )    g.setColor(ColorGood);
			else if( NTR-SR<+1000000 ) g.setColor(ColorNear);
			else                       g.setColor(ColorFar);
		} else {
			int d = (int)Math.sqrt((NTX-SX)*(NTX-SX)+(NTY-SY)*(NTY-SY));
			if( d<+1000 )    g.setColor(ColorGood);
			else if( d<+1000000 ) g.setColor(ColorNear);
			else                      g.setColor(ColorFar);
		}
		drawDot(g, SX, SY, 10);
		if(MODE<=2)
			g.drawString(String.format("R: %10d", (int)(NTR-SR)), 0, 30);
		if(FROMME)
			g.drawLine((int)(300-20*N_VX), (int)(300+20*N_VY), (int)(300), (int)(300));
		else
			g.drawLine((int)(300+SX*ra-20*N_VX), (int)(300-SY*ra+20*N_VY), (int)(300+SX*ra), (int)(300-SY*ra));

		if(MODE>=2) {
			int d = (int)Math.sqrt((NTX-SX)*(NTX-SX)+(NTY-SY)*(NTY-SY));
			if( d<+1000 )    g.setColor(ColorGood);
			else if( d<+1000000 ) g.setColor(ColorNear);
			else                      g.setColor(ColorFar);
			g.drawString(String.format("S-T: %10d", d), 400, 30);
		}
	}
}
