package VM;
import java.io.*;

public class LoggingVM implements VM, Cloneable, Serializable
{
	private VM           impl;
	transient private OutputStream os;

	public LoggingVM( VM impl, String logFileName ) throws IOException, FileNotFoundException
	{
		this.impl = impl;
		this.os = new FileOutputStream(logFileName);
		writeLE( 0xCAFEBABE );
		writeLE( 346 );
		writeLE( impl.getConfigID() );

		Runtime.getRuntime().addShutdownHook(new Thread(){
			public void run() { try{close();}catch(Exception e){} }
		});
	}

	public int getTimeStamp()    { return impl.getTimeStamp(); }
	public int getConfigID()     { return impl.getConfigID(); }
	public VM snapshot()         { return impl.snapshot(); }
	public double readPort(int i){ return impl.readPort(i); }

	private double px=-1, py=-1, pc=-1;
	public void step( double dvx, double dvy ) throws IOException
	{
		double cnf = impl.getConfigID();

		int cnt = (px!=dvx?1:0) + (py!=dvy?1:0) + (pc!=cnf?1:0);
		if(cnt>0) {writeLE(impl.getTimeStamp()); writeLE(cnt);}
		if(px!=dvx) { writeLE(2); writeLE(dvx); px=dvx; }
		if(py!=dvy) { writeLE(3); writeLE(dvy); py=dvy; }
		if(pc!=cnf) { writeLE(0x3E80); writeLE(cnf); pc=cnf; }

		impl.step(dvx, dvy);
	}


	private void close() throws IOException
	{
		if( os != null )
		{
			writeLE(impl.getTimeStamp());
			writeLE(0);
			os.close();
			os = null;
		}
	}

	private void writeLE( int n ) throws IOException
	{
		if( os != null )
			for(int i=0; i<32; i+=8)
				os.write( (n>>i)&0xff );
	}

	private void writeLE( double d ) throws IOException
	{
		long n = Double.doubleToLongBits(d);
		writeLE( (int)(n&0xffffffff) );
		writeLE( (int)((n>>32)&0xffffffff) );
	}
}
