#include <iostream>
#include <cstdio>
#include <vector>
#include <cassert>
#include <set>
#include <algorithm>
#include <fstream>
#include <sstream>
using namespace std;

vector<unsigned char> read_file( const char* filename )
{
	FILE* fp = fopen(filename, "rb");

	fseek(fp, 0, SEEK_END);
	size_t size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	vector<unsigned char> buf(size);
	fread(&buf[0], 1, size, fp);
	fclose(fp);

	return buf;
}

template <typename T>
T raw_read( void* p )
{
	return *static_cast<T*>(p);
}

struct Optimizer
{
	vector<unsigned char>& bin;
	int N;
	Optimizer( vector<unsigned char>& bin ) : bin(bin), N(bin.size()/12) {}

	multiset<int> read_cell;
	bool every_cmp_followed_by_phi;

	void inspect()
	{
		every_cmp_followed_by_phi = true;
		for(int rd=0; rd<N; ++rd)
		{
			unsigned int ins = raw_read<unsigned int>(rd%2==0 ? &bin[rd*12+8] : &bin[rd*12]);
			unsigned int OP   = (ins>>28)&0xf;
			unsigned int r1   = (ins>>14)&0x3fff;
			unsigned int r2   = ins&0x3fff;
			unsigned int S_OP = (r1>>10)&0xf;
			unsigned int IMM  = r1&0x3ff;
			unsigned int C_OP = (IMM>>7)&0x7;
			if( OP!=0 && OP!=5 ) read_cell.insert(r1), read_cell.insert(r2);
			if(OP==5) read_cell.insert(r2);
			if(OP==0) read_cell.insert(r2);

			if( (OP==0&&S_OP==1) && !(rd+1<N &&
				((raw_read<unsigned int>((rd+1)%2==0 ? &bin[(rd+1)*12+8] : &bin[(rd+1)*12])>>28)&0xf)==6) )
				every_cmp_followed_by_phi = false;
		}
	}
};

int main( int argc, const char* argv[] )
{
	if( argc < 3 )
	{
		cerr << "Usage:\n" << endl;
		cerr << "  create_server bin1.obf VM/VM_template.java VM1 > VM1.java\n" << endl;
		return -1;
	}

	const char* obf_file  = argv[1];
	const char* tmpl_file = argv[2];
	string CLASSNAME = argv[3];
	bool OPTIMIZE = true;

	vector<unsigned char> bin = read_file( argv[1] );
	const int N = bin.size() / 12;
	Optimizer opt(bin);
	opt.inspect();

// vm-main --------------------------------------------------------------------

	string VMINIT;
	{
		stringstream ss;
		for(int rd=0; rd<N; ++rd)
		{
			double v = raw_read<double>(rd%2==0 ? &bin[rd*12] : &bin[rd*12+4]);
			if( !OPTIMIZE || opt.read_cell.count(rd) )
			{
				char buf[255];
				sprintf(buf, "\t\tmem[%d] = %.32A; // %lf", rd, v, v);
				ss << buf << endl;
			}
		}
		VMINIT = ss.str();
	}

	string VMINIT2;
	{
		stringstream ss;
		for(int rd=0; rd<N; ++rd)
		{
			double v = raw_read<double>(rd%2==0 ? &bin[rd*12] : &bin[rd*12+4]);
			if( !OPTIMIZE || opt.read_cell.count(rd) )
			{
				char buf[255];
				sprintf(buf, "\tdouble m%d = %.32A; // %lf", rd, v, v);
				ss << buf << endl;
			}
		}
		VMINIT2 = ss.str();
	}

	string VMMAIN;
	{
		stringstream ss;
		for(int rd=0; rd<N; ++rd)
		{
			unsigned int ins = raw_read<unsigned int>(rd%2==0 ? &bin[rd*12+8] : &bin[rd*12]);
			unsigned int OP   = (ins>>28)&0xf;
			unsigned int r1   = (ins>>14)&0x3fff;
			unsigned int r2   = ins&0x3fff;
			unsigned int S_OP = (r1>>10)&0xf;
			unsigned int IMM  = r1&0x3ff;
			unsigned int C_OP = (IMM>>7)&0x7;

			switch(OP)
			{
			default:
				assert(0);
			case 1:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tmem["<<rd<<"] = mem["<<r1<<"] + mem["<<r2<<"];" << endl;
				break;
			case 2:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tmem["<<rd<<"] = mem["<<r1<<"] - mem["<<r2<<"];" << endl;
				break;
			case 3:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tmem["<<rd<<"] = mem["<<r1<<"] * mem["<<r2<<"];" << endl;
				break;
			case 4:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tmem["<<rd<<"] = mem["<<r1<<"]==0 ? 0 : mem["<<r1<<"] / mem["<<r2<<"];" << endl;
				break;
			case 5:
				ss << "\t\tport_out["<<r1<<"] = mem["<<r2<<"];" << endl;
				break;
			case 6:
				if(!OPTIMIZE || !opt.every_cmp_followed_by_phi)
				{
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tmem["<<rd<<"] = mem[status ? "<<r1<<" : "<<r2<<"];" << endl;
				}
				else
				{
					ss << " mem["<<rd<<"]=mem["<<r1<<"]; else mem["<<rd<<"]=mem["<<r2<<"];" << endl;
				}
				break;
			case 0:
				switch(S_OP)
				{
				default:
					assert(0);
				case 0:
					break;
				case 1:
					if(!OPTIMIZE || !opt.every_cmp_followed_by_phi)
						switch(C_OP)
						{
						default:
							assert(0);
						case 0:
							ss << "\t\tstatus = (mem["<<r2<<"] < 0);" << endl;
							break;
						case 1:
							ss << "\t\tstatus = (mem["<<r2<<"] <= 0);" << endl;
							break;
						case 2:
							ss << "\t\tstatus = (mem["<<r2<<"] == 0);" << endl;
							break;
						case 3:
							ss << "\t\tstatus = (mem["<<r2<<"] => 0);" << endl;
							break;
						case 4:
							ss << "\t\tstatus = (mem["<<r2<<"] > 0);" << endl;
							break;
						}
					else
						switch(C_OP)
						{
						default:
							assert(0);
						case 0:
							ss << "\t\tif(mem["<<r2<<"] < 0)";
							break;
						case 1:
							ss << "\t\tif(mem["<<r2<<"] <= 0)";
							break;
						case 2:
							ss << "\t\tif(mem["<<r2<<"] == 0)";
							break;
						case 3:
							ss << "\t\tif(mem["<<r2<<"] => 0)";
							break;
						case 4:
							ss << "\t\tif(mem["<<r2<<"] > 0)";
							break;
						}
				case 2:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tmem["<<rd<<"] = Math.sqrt(mem["<<r2<<"]);" << endl;
					break;
				case 3:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tmem["<<rd<<"] = mem["<<r2<<"];" << endl;
					break;
				case 4:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tmem["<<rd<<"] = port_in["<<r2<<"];" << endl;
					break;
				}
			}
		}
		VMMAIN = ss.str();
	}

	string VMMAIN2;
	{
		stringstream ss;
		for(int rd=0; rd<N; ++rd)
		{
			unsigned int ins = raw_read<unsigned int>(rd%2==0 ? &bin[rd*12+8] : &bin[rd*12]);
			unsigned int OP   = (ins>>28)&0xf;
			unsigned int r1   = (ins>>14)&0x3fff;
			unsigned int r2   = ins&0x3fff;
			unsigned int S_OP = (r1>>10)&0xf;
			unsigned int IMM  = r1&0x3ff;
			unsigned int C_OP = (IMM>>7)&0x7;

			switch(OP)
			{
			default:
				assert(0);
			case 1:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tm"<<rd<<" = m"<<r1<<" + m"<<r2<<";" << endl;
				break;
			case 2:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tm"<<rd<<" = m"<<r1<<" - m"<<r2<<";" << endl;
				break;
			case 3:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tm"<<rd<<" = m"<<r1<<" * m"<<r2<<";" << endl;
				break;
			case 4:
				if( !OPTIMIZE || opt.read_cell.count(rd) )
					ss << "\t\tm"<<rd<<" = m"<<r1<<"==0 ? 0 : m"<<r1<<" / m"<<r2<<";" << endl;
				break;
			case 5:
				ss << "\t\tport_out["<<r1<<"] = m"<<r2<<";" << endl;
				break;
			case 6:
				if(!OPTIMIZE || !opt.every_cmp_followed_by_phi)
				{
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tm"<<rd<<" = mstatus ? "<<r1<<" : "<<r2<<";" << endl;
				}
				else
				{
					ss << " m"<<rd<<"=m"<<r1<<"; else m"<<rd<<"=m"<<r2<<";" << endl;
				}
				break;
			case 0:
				switch(S_OP)
				{
				default:
					assert(0);
				case 0:
					break;
				case 1:
					if(!OPTIMIZE || !opt.every_cmp_followed_by_phi)
						switch(C_OP)
						{
						default:
							assert(0);
						case 0:
							ss << "\t\tstatus = (m"<<r2<<" < 0);" << endl;
							break;
						case 1:
							ss << "\t\tstatus = (m"<<r2<<" <= 0);" << endl;
							break;
						case 2:
							ss << "\t\tstatus = (m"<<r2<<" == 0);" << endl;
							break;
						case 3:
							ss << "\t\tstatus = (m"<<r2<<" => 0);" << endl;
							break;
						case 4:
							ss << "\t\tstatus = (m"<<r2<<" > 0);" << endl;
							break;
						}
					else
						switch(C_OP)
						{
						default:
							assert(0);
						case 0:
							ss << "\t\tif(m"<<r2<<" < 0)";
							break;
						case 1:
							ss << "\t\tif(m"<<r2<<" <= 0)";
							break;
						case 2:
							ss << "\t\tif(m"<<r2<<" == 0)";
							break;
						case 3:
							ss << "\t\tif(m"<<r2<<" => 0)";
							break;
						case 4:
							ss << "\t\tif(m"<<r2<<" > 0)";
							break;
						}
				case 2:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tm"<<rd<<" = Math.sqrt(m"<<r2<<");" << endl;
					break;
				case 3:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tm"<<rd<<" = m"<<r2<<";" << endl;
					break;
				case 4:
					if( !OPTIMIZE || opt.read_cell.count(rd) )
						ss << "\t\tm"<<rd<<" = port_in["<<r2<<"];" << endl;
					break;
				}
			}
		}
		VMMAIN2 = ss.str();
	}

// output --------------------------------------------------------------------

	ifstream fin(tmpl_file);
	for(string line; getline(fin, line); )
	{
		size_t i;
		if( (i=line.find("<%VMINIT%>")) != string::npos )
			cout << VMINIT;
		else if( (i=line.find("<%VMMAIN%>")) != string::npos )
			cout << VMMAIN;
		else if( (i=line.find("<%VMINIT2%>")) != string::npos )
			cout << VMINIT2;
		else if( (i=line.find("<%VMMAIN2%>")) != string::npos )
			cout << VMMAIN2;
		else
		{
			while( (i=line.find("<%CLASSNAME%>")) != string::npos )
				line.replace(i, 13, CLASSNAME);
			cout << line << endl;
		}
	}
}
